/*  (c) Copyright 2000, Andrew Rambaut                 */
/*  Department of Zoology, University of Oxford        */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#include <time.h>

#ifdef __MWERKS__
#include <console.h>
#endif	

#include "lik.h"
#include "calclik.h"
#include "maxlik.h"
#include "evolve.h"
#include "tree.h"
#include "fF84.h"
#include "gamma.h"
#include "search.h"
#include "random.h"
#include "progress.h"

#define PROGRAM_NAME "shtests"
#define VERSION_NUMBER "Version 1.0"

char *ModelNames[numModels]={
	"F84",
	"HKY",
	"REV"
};

int printHelp, writeDeltas, treeFile, testKind;
char treeFileName[256];
FILE *tree_fv;

enum {
	RELL,
	FULL,
	PARA,
	numTestKinds
};

char *TestKindNames[numTestKinds]={
	"RELL",
	"FULL",
	"PARA"
};

/* prototypes */

static void PrintTitle();
static void PrintUsage();
void PrintVerbose(TTree *tree);
void ReadParams();
void ReadFileParams();
void ReadFile();
int OpenTreeFile();
TTree *ReadNextTree(TTree *tree);
void CloseTreeFile();

void PrintSiteLik(FILE *fv, double *siteLikelihoods);
void StoreSiteLik(double *siteLikelihoods);
void WriteSequences(FILE *fv, TTree *tree);
void PerformRELLTest(double *treeLikelihoods, double **siteLikelihoods);
void PerformFULLTest(TTree **trees, double *treeLikelihoods);
void PerformPARATest(TTree **trees, double *treeLikelihoods);

/* functions */

static void PrintTitle()
{
	fprintf(stderr, "SH Test of phylogenetic hypotheses - SHTests\n");
	fprintf(stderr, "%s\n", VERSION_NUMBER);
	fprintf(stderr, "(c) Copyright, 2000 Andrew Rambaut\n");
	fprintf(stderr, "Department of Zoology, University of Oxford\n");
	fprintf(stderr, "South Parks Road, Oxford OX1 3PS, U.K.\n\n");
}

static void PrintUsage()
{
	fprintf(stderr, "Usage: %s [-mMODEL] [-p nnn] [+c|-c #1 #2 [#3]] [-e] [-g #] [-a #] [-n #] [-kKIND] [-h|-help] [parameters]\n          [-u[l]] [-v[b][m][p][s][w]|-q] [treefile]\n", PROGRAM_NAME);
	fprintf(stderr, "  -m: MODEL = F84, HKY or REV [default=HKY]\n");
	fprintf(stderr, "  -p: CODON CATEGORIES = 112, 123, 120, etc. [homogeneity]\n");
	fprintf(stderr, "  -c: #1 #2 #3 = codon category rates [default = estimate rates].\n");
	fprintf(stderr, "  -e: estimate seperate models for each site category [default = 1 model].\n");
	fprintf(stderr, "  -g: # = num categories for gamma rate heterogeneity [default = homogeneity].\n");
	fprintf(stderr, "  -a: # = gamma shape (alpha) for site rate heterogeneity [default = estimate].\n");
	fprintf(stderr, "  -n: # = number of bootstrap replicates [default = 1000].\n");
	fprintf(stderr, "  -k: KIND = RELL or FULL [default = RELL].\n");
/*	fprintf(stderr, "  -k: KIND = RELL, FULL or PARAmetric [default = RELL].\n");*/
	fprintf(stderr, "  -h,-help = Give this help message (with details of selected model)\n");
	fprintf(stderr, "  parameters = Specific to model.  To get details use: %s -mMODEL -help \n", PROGRAM_NAME);
	fprintf(stderr, "  -u: Use user trees, -ul Use user branch lengths\n");
	fprintf(stderr, "  -v: Verbose information\n");
	fprintf(stderr, "    b Print final branch lengths\n");
	fprintf(stderr, "    m Memory usage information\n");
	fprintf(stderr, "    p Show indication of progress\n");
	fprintf(stderr, "    s Write likelihoods for each Site\n");
	fprintf(stderr, "    d Write deltas for each replicate\n");
	fprintf(stderr, "    w Write trees [default=don't]\n");
	fprintf(stderr, "  -q: Quiet\n");
	fprintf(stderr, "  treefile: Optional file containing trees in PHYLIP format\n");
}


void ReadParams(int argc, char **argv)
{
	int i, j, estimate, modelPerCat;
	char ch;
	char *P, st[4];
	
	model=HKY;
	printHelp=0;
	treeFile=0;
	numDatasets=1;
	
	userTrees=1;
	userLengths=0;
	writeTrees=0;
	writeDeltas=0;
	verbose=0;
	verboseMemory=0;
	verboseLengths=0;
	verboseSiteLik=0;
	showProgress=0;
	quiet=0;

	estimateGlobals=1;

	debug=0;
	debugSplits=0;
	debugTrees=0;
	
	numCategories=1;
	numModelCats=1;
	modelPerCat=0;
	codonCats[0]=codonCats[1]=codonCats[2]=0;
	for (i=0; i<MAX_CATEGORIES; i++)
		siteRate[i]=1.0;
	codonRates=0;
	estimateRates=1;
	gammaRates=0;
	alpha=1.0;
	estimateAlpha=1;
	
	molClock=0;
	datedTips=0;
	estimateAbsRate=0;
	absRate=0.001;

	estimateRoot=0;
	rootTaxon=-1;

	estimateClocks=0;
	estimateClockPos=0;
	numClocks=1;
	
	estimateIntervals=0;
	intervalLimits=1.92;

	numReplicates=1000;
	testKind = 0; /* 0 = RELL, 1 = FULL, 2 = PARAMETRIC */
	
	i=1; 
	while (i<argc) {
		P=argv[i];
		if (P[0]=='-') {
			ch=toupper(P[1]);
			switch (ch) {
				case 'H':
					printHelp=1;
				break;
				case 'Q':
					quiet=1;
				break;
				case 'M':
					if (GetStrParam(argc, argv, &i, P+2, st, 3)) {
						fprintf(stderr, "Bad (or missing) Model Code: %s\n\n", argv[i]);
						PrintUsage();
						exit(0);
					}
					
					P=st;
					model=-1;
					for (j=F84; j<numModels; j++) {
						if (strncmp(P, ModelNames[j], 3)==0)
							model=j;
					}
					if (model==-1) {
						fprintf(stderr, "Unknown Model: %s\n\n", argv[i]);
						PrintUsage();
						exit(0);
					}
				break;
			}
		}
		i++;
	}

	if (!quiet || printHelp)
 		PrintTitle();
	if (printHelp)
 		PrintUsage();
	SetModel(printHelp);
	if (printHelp)
		exit(0);
	
	i=1; 
	while (i<argc) {
		P=argv[i];
		if (*P!='-' && *P!='+') {
			if (treeFile) {
				fprintf(stderr, "Illegal command parameter: %s\n\n", argv[i]);
				PrintUsage();
				exit(0);
			}
			treeFile=1;
			strcpy(treeFileName, argv[i]);
		} else {
			while (*P) {
				*P=toupper(*P);
				P++;
			}
			P=argv[i];
			if (*P=='+')
				estimate=1;
			else 
				estimate=0;
			P++;
			ch=*P;
			P++;
			switch (ch) {
				case 'M': 	/* although this has already been read, we need to
							   use it up in this run through of the params.    */
					GetStrParam(argc, argv, &i, P, st, 3);
				break;
				case 'H':
				case 'Q':
				break;
				case 'P':
					if (GetStrParam(argc, argv, &i, P, st, 3)) {
						fprintf(stderr, "Bad (or missing) Codon Categories Code: %s\n\n", argv[i]);
						PrintUsage();
						exit(0);
					}
					P=st;
					if (strlen(P)==3) {
						int cp1=0, cp2=0, cp3=0;
						
						for (j=0; j<3; j++) {
							codonCats[j]=*P-'1';
							switch (codonCats[j]) {
								case -1: break;  /* exclude this codon position */
								case 0: cp1=1; break;
								case 1: cp2=1; break;
								case 2: cp3=1; break;
								default:
									fprintf(stderr, "Codon categories should be from 0-3: %s\n\n", argv[i]);
									PrintUsage();
									exit(0);
							}
							P++;
						}
						if (!cp1 || (cp3 && !cp2)) {
							fprintf(stderr, "Codon categories should be from 0-3: %s\n\n", argv[i]);
							PrintUsage();
							exit(0);
						}
						numCategories=cp1+cp2+cp3;
						codonRates=1;
					} else {
						fprintf(stderr, "Must have a category for each codon position: %s\n\n", argv[i]);
						PrintUsage();
						exit(0);
					}
				break;
				case 'C':
					estimateRates=0;
					if (numCategories==3) {
						if (GetDoubleParams(argc, argv, &i, P, 3, siteRate)) {
							fprintf(stderr, "Bad Codon Category Rates: %s\n\n", argv[i]);
							PrintUsage();
							exit(0);
						}
					} else if (numCategories==2) {
						if (GetDoubleParams(argc, argv, &i, P, 2, siteRate)) {
							fprintf(stderr, "Bad Codon Category Rate: %s\n\n", argv[i]);
							PrintUsage();
							exit(0);
						}
					} else {
						fprintf(stderr, "Must have two or three codon categories set to specify rates: %s\n\n", argv[i]);
						PrintUsage();
						exit(0);
					}
				break;
				case 'E':
					modelPerCat=1;
				break;
				case 'N':
					if (GetIntParams(argc, argv, &i, P, 1, &numReplicates) || numReplicates<2) {
						fprintf(stderr, "Bad number of bootstrap replicates: %s\n\n", argv[i]);
						PrintUsage();
						exit(0);
					}
				break;
				case 'K':
					if (GetStrParam(argc, argv, &i, P, st, 4)) {
						fprintf(stderr, "Bad (or missing) Kind Code: %s\n\n", argv[i]);
						PrintUsage();
						exit(0);
					}
					
					P=st;
					testKind=-1;
					for (j=RELL; j<numTestKinds; j++) {
						if (strncmp(P, TestKindNames[j], 4)==0)
							testKind=j;
					}
					if (testKind==-1) {
						fprintf(stderr, "Unknown Test Kind: %s\n\n", argv[i]);
						PrintUsage();
						exit(0);
					}
				break;
				case 'G':
					if (GetIntParams(argc, argv, &i, P, 1, &numCategories) || numCategories<1 || numCategories>MAX_CATEGORIES) {
						fprintf(stderr, "Bad number of Gamma rate categories: %s\n\n", argv[i]);
						PrintUsage();
						exit(0);
					}
					gammaRates=1;
				break;
				case 'A':
					if (gammaRates) {
						if (GetDoubleParams(argc, argv, &i, P, 1, &alpha) || alpha<0.0) {
							fprintf(stderr, "Bad alpha shape for Gamma rate heterogeneity: %s\n\n", argv[i]);
							PrintUsage();
							exit(0);
						}
						estimateAlpha=0;
					} else {
						fprintf(stderr, "Must specify the number of gamma rate categories before setting alpha: %s\n\n", argv[i]);
						PrintUsage();
						exit(0);
					}
				break;
				case 'U':
					userTrees=1;
					if (*P=='L')
						userLengths=1;
				break;
				case 'I':
					estimateIntervals=1;
					intervalLimits=1.92;
				break;
				case 'V':
					verbose=1;
					while (*P!='\0') {
						switch (*P) {
							case 'P': showProgress=1; break;
							case 'M': verboseMemory=1; break;
							case 'B': verboseLengths=1; break;
							case 'S': verboseSiteLik=1; break;
							case 'W': writeTrees=1; break;
							case 'D': writeDeltas=1; break;
							default:					
								fprintf(stderr, "Unknown verbose mode: %s\n\n", argv[i]);
								PrintUsage();
								exit(0);
						}
						P++;
					}
				break;
				case 'D':
					debug=1;
					while (*P!='\0') {
						switch (*P) {
							case 'S': debugSplits=1; break;
							case 'T': debugTrees=1; break;
							case 'W': writeTrees=1; break;
							default:					
								fprintf(stderr, "Unknown debug mode: %s\n\n", argv[i]);
								PrintUsage();
								exit(0);
						}
						P++;
					}
				break;
				default:
					if (ReadModelParam(argc, argv, &i)==-1) {
						PrintUsage();
						exit(0);
					}
				break;
			}
		}
		i++;
	}
	
	if (numCategories<2) {
		estimateRates=0;
		estimateAlpha=0;
		gammaRates=0;
	}	
	
	if (gammaRates)
		estimateRates=0;
		
	if (modelPerCat) {
		if (gammaRates) {
			fprintf(stderr, "You can't have a gamma model for each category (at the moment)\n\n");
			PrintUsage();
			exit(0);
		}
		numModelCats=numCategories;
	}
	
	if (FinishModelParam()<0) {
		PrintUsage();
		exit(0);
	}
}


void ReadFileParams()
{
	char ch, st[256];
	
	ch=fgetc(stdin);
	while (!feof(stdin) && isspace(ch)) 
		ch=fgetc(stdin);
	ungetc(ch, stdin);
	fgets(st, 255, stdin);
	if ( sscanf( st, " %d %d", &numTaxa, &numBases)!=2 ) {
		fprintf(stderr, "Unable to read parameters from standard input\n");
		exit(0);
	}
}


void ReadFile()
{
	int n, b, i;
	char ch;
		
	n=0;
	do {
		ch=fgetc(stdin);
		while ( !feof(stdin) && isspace(ch)) 
			ch=fgetc(stdin);
			
		if ( feof(stdin) ) {
			fprintf(stderr, "Unexpected end of file on standard input\n"); 
			exit(0);
		}
	
		i=0;
		while ( i<MAX_NAME_LEN && !feof(stdin) && !isspace(ch) ) {
			names[n][i]=ch;
			ch=fgetc(stdin);
			i++;
		}
		names[n][i]='\0';
		if (i==0) {
			fprintf(stderr, "Name missing for species %d\n", n+1);
			exit(0);
		}
		while (!feof(stdin) && isspace(ch))
			ch=fgetc(stdin);
		
		if ( feof(stdin) ) {
			fprintf(stderr, "Unexpected end of file on standard input\n");
			exit(0);
		}
		
		b=0;
		while ( !feof(stdin) && b<numBases) {
			if ( !isspace(ch) ) {
				sequences[n][b]=ch;
				b++;
			}
			ch=toupper(fgetc(stdin));
		}
		
		if ( b<numBases ) {
			fprintf(stderr, "Unexpected end of file on standard input\n");
			exit(0);
		}
		
		/*fprintf(stderr, "%d: %s, bases read: %d\n", n+1, names[n], b);*/
		n++;
		
		if ( n<numTaxa && feof(stdin) ) {
			fprintf(stderr, "Too few sequences in input file\n");
			exit(0);
		}
	} while ( n<numTaxa );
}


int OpenTreeFile()
{
	char st[256];
	int n;
		
	if (treeFile) {
		if ( (tree_fv=fopen(treeFileName, "rt"))==NULL ) {
			fprintf(stderr, "Error opening tree file: '%s'\n", treeFileName);
			exit(0);
		}
		n=0;
		while (!feof(tree_fv)) {
			if (fgetc(tree_fv)==';')
				n++;
		}
		rewind(tree_fv);
	} else {
		tree_fv=NULL;
		fgets(st, 255, stdin);
		if ( sscanf(st, " %d ", &n)!=1 ) {
			fprintf(stderr, "Tree is missing from end of sequence file\n");
			exit(0);
		}
	}
	
	return n;
}

TTree *ReadNextTree(TTree *tree)
{		
	if (treeFile) {
		ReadTree(tree_fv, tree);
	} else {
		ReadTree(stdin, tree);
	}
	
	return tree;
}

void CloseTreeFile()
{
	if (treeFile)
		fclose(tree_fv);
}

void PrintVerbose(TTree *tree)
{
	int i;
	
	fprintf(stderr, "Performing SH test, %s version\n", TestKindNames[testKind]);
	fprintf(stderr, "  %d bootstrap replicates\n\n", numReplicates);

	fprintf(stderr, "SHIMODAIRA, H., AND M. HASEGAWA. 1999. Multiple comparisons of\n");
	fprintf(stderr, "  log-likelihoods with applications to phylogenetic inference.\n");
	fprintf(stderr, "  Mol. Biol. Evol. 16:1114 1116.\n\n");

	fprintf(stderr, "%d taxa, %d bases\n", numTaxa, numBases);
	fprintf(stderr, "	%d distinct patterns\n\n", numPatterns);
	if (userTrees)
		if (numTrees==1)
			fprintf(stderr, "Calculating likelihood of user supplied tree,\n");
		else
			fprintf(stderr, "Calculating likelihood of %d user supplied trees,\n", numTrees);
	if (userLengths)
		fprintf(stderr, "  using branch lengths of supplied tree(s).\n");
	else {
		fprintf(stderr, "  iterating to find maximum likelihood branch lengths.\n");
	}
	if (numCategories>1) {
		fprintf(stderr, "Rate heterogeneity:\n");
		if (gammaRates) {
			fprintf(stderr, "  Discrete Gamma rate heterogeneity with %d categories\n", numCategories);
			if (estimateAlpha)
				fprintf(stderr, "  estimating shape parameter, alpha.\n");
			else
				fprintf(stderr, "  shape parameter, alpha = %f\n", alpha);
		} else {
			fprintf(stderr, "  Number of codon position categories = %d, pattern = %d%d%d\n", 
							numCategories, codonCats[0]+1, codonCats[1]+1, codonCats[2]+1);
			if (estimateRates)
				fprintf(stderr, "  estimating ML rates for categories");
			else
				fprintf(stderr, "  rates for categories supplied by user\n");
		}
		if (!estimateAlpha && !estimateRates) {
			fprintf(stderr, "    rates =");
			for (i=0; i<numCategories; i++)
				fprintf(stderr, " %f", siteRate[i]);
			fputc('\n', stderr);
		}
	} else
		fprintf(stderr, "Rate homogeneity assumed");
	fprintf(stderr, ".\n");
	fprintf(stderr, "Model=%s\n\n", ModelNames[model]);
}
	
	
void PrintSiteLik(FILE *fv, double *siteLikelihoods)
{
	short *T;
	int k;

	T=patternIndex;
	fprintf(fv, "%lf", siteLikelihoods[0]);
	for(k = 1; k < numBases; k++)
		fprintf(fv, "\t%lf", siteLikelihoods[k]);
	fprintf(fv, "\n");
}

void StoreSiteLik(double *siteLikelihoods)
{
	short *T;
	long k;

	T=patternIndex;
	for (k = 0; k < numBases; k++) {
		siteLikelihoods[k] = patternLik[0][*T];
		T++;
	}
}

void WriteSequences(FILE *fv, TTree *tree)
{
	int i, j;
	char *P;

	fprintf(fv, " %d %d\n", numTaxa, numBases);

	for (i=0; i<numTaxa; i++) {
		fprintf(fv, "%s ", names[i]);
		P=sequences[i];
		for (j=0; j<numBases; j++) {
			fputc(*P, fv);
			P++;
		}
		fputc('\n', fv);
	}
}


void PerformRELLTest(double *treeLikelihoods, double **siteLikelihoods)
{
	double **likelihoods;
	double *meanLik;
	double maxLik, bestLik, probTree;
	long n, i, j, r, maxTree;
	
	meanLik = (double *)AllocMem(numTrees*sizeof(double), "meanLik", "PerformRELLTest", 1);
	likelihoods = (double **)AllocMem(numTrees*sizeof(double*), "likelihoods", "PerformRELLTest", 1);
	for (j = 0; j < numTrees; j++) {
		likelihoods[j] = (double *)AllocMem(numReplicates*sizeof(double), "likelihoods[]", "PerformRELLTest", 1);
	}
	
	/* Find the maximum likelihood tree */
	maxLik = treeLikelihoods[0];
	maxTree = 0;
	for (j = 1; j < numTrees; j++) {
		if (treeLikelihoods[j] > maxLik) {
			maxLik = treeLikelihoods[j];
			maxTree = j;
		}
	}
	bestLik = maxLik;

	/* Calculate the deltas for each tree */
	for (j = 0; j < numTrees; j++) {
		treeLikelihoods[j] = maxLik - treeLikelihoods[j];
		meanLik[j] = 0.0;
	}

	if (!quiet) {
		fprintf(stderr, "Generating bootstrap replicates...\n");
		InitProgressBar(numReplicates);
		DrawProgressBar();
	}

	/* For each bootstrap replicate... */
	for (n = 0; n < numReplicates; n++) {
		for (j = 0; j < numTrees; j++)
			likelihoods[j][n] = 0.0;
			
		for (i = 0; i < numBases; i++) {
			r = (long)(rndu() * numBases);
			for (j = 0; j < numTrees; j++)
				likelihoods[j][n] += siteLikelihoods[j][r];
		}

		for (j = 0; j < numTrees; j++)
			meanLik[j] += likelihoods[j][n];
			
		if (!quiet)
			ProgressBar();
	}
	
	/* Centre the resampled likelihoods and find the maximum */
	for (j = 0; j < numTrees; j++)
		meanLik[j] /= numReplicates;

	for (n = 0; n < numReplicates; n++) {
		likelihoods[0][n] -= meanLik[0];
		maxLik = likelihoods[0][n];
		for (j = 1; j < numTrees; j++) {
			likelihoods[j][n] -= meanLik[j];
			if (likelihoods[j][n] > maxLik)
				maxLik = likelihoods[j][n];
		}

	/* Calculate the distribution of deltas for this replicate */
		for (j = 0; j < numTrees; j++)
			likelihoods[j][n] = maxLik - likelihoods[j][n];
	}

	if (writeDeltas) {
		fprintf(stdout, "Delta:");
		for (j = 0; j < numTrees; j++)
			fprintf(stdout, "\t%lf", treeLikelihoods[j]);
		fprintf(stdout, "\n\nRep");
		for (j = 0; j < numTrees; j++)
			fprintf(stdout, "\tTree%d", j+1);
		for (n = 0; n < numReplicates; n++) {
			fprintf(stdout, "%d", n+1);
			for (j = 0; j < numTrees; j++)
				fprintf(stdout, "\t%lf", likelihoods[j][n]);
			fprintf(stdout, "\n");
		}
		fprintf(stdout, "\n");
	}
	
	fprintf(stdout, "\tTree\tlnL\tdelta\tP(delta)\n");
	for (j = 0; j < numTrees; j++) {
		if (j == maxTree)
			fprintf(stdout, "ML");

		probTree = 0.0;
		for (n = 0; n < numReplicates; n++) {
			if (treeLikelihoods[j] <= likelihoods[j][n])
				probTree++;
		}
		probTree /= numReplicates;
		
		fprintf(stdout, "\t%d\t%lf\t%lf\t%lf\n", j+1, 
			bestLik - treeLikelihoods[j],
			treeLikelihoods[j], probTree);
	}
}

void PerformFULLTest(TTree **trees, double *treeLikelihoods)
{
	double **likelihoods;
	char **originalSequences;
	double *meanLik;
	double maxLik, bestLik, probTree;
	long n, i, j, r, maxTree;
	
	meanLik = (double *)AllocMem(numTrees*sizeof(double), "meanLik", "PerformFULLTest", 1);

	likelihoods = (double **)AllocMem(numTrees*sizeof(double*), "likelihoods", "PerformFULLTest", 1);
	for (j = 0; j < numTrees; j++) {
		likelihoods[j] = (double *)AllocMem(numReplicates*sizeof(double), "likelihoods[]", "PerformFULLTest", 1);
	}

	originalSequences = sequences;
	sequences = (char **)AllocMem(numTaxa*sizeof(char*), "sequences", "PerformFULLTest", 1);
	for (j = 0; j < numTaxa; j++) {
		sequences[j] = (char *)AllocMem(numBases*sizeof(char), "sequences[]", "PerformFULLTest", 1);
	}
	
	/* Find the maximum likelihood tree */
	maxLik = treeLikelihoods[0];
	maxTree = 0;
	for (j = 1; j < numTrees; j++) {
		if (treeLikelihoods[j] > maxLik) {
			maxLik = treeLikelihoods[j];
			maxTree = j;
		}
	}

	bestLik = maxLik;

	/* Calculate the deltas for each tree */
	for (j = 0; j < numTrees; j++) {
		treeLikelihoods[j] = maxLik - treeLikelihoods[j];
		meanLik[j] = 0.0;
	}

	if (!quiet) {
		fprintf(stderr, "Generating bootstrap replicates...\n");
		InitProgressBar(numReplicates);
		DrawProgressBar();
	}

	/* For each bootstrap replicate... */
	for (n = 0; n < numReplicates; n++) {
		for (i = 0; i < numBases; i++) {
			r = (long)(rndu() * numBases);
			for (j = 0; j < numTaxa; j++)
				sequences[j][i] = originalSequences[j][r];
		}

		CreatePatterns();
		
		for (j = 0; j < numTrees; j++) {
			SetTipLikelihoods(trees[j]);
			quiet = 1;
			SetupLikelihood(trees[j]);
			quiet = 0;

			TreeLik(trees[j]);
			likelihoods[j][n] = trees[j]->lnL;
			meanLik[j] += trees[j]->lnL;
		}
			
		if (!quiet)
			ProgressBar();
	}
	
	/* Centre the resampled likelihoods and find the maximum */
	for (j = 0; j < numTrees; j++)
		meanLik[j] /= numReplicates;

	for (n = 0; n < numReplicates; n++) {
		likelihoods[0][n] -= meanLik[0];
		maxLik = likelihoods[0][n];
		for (j = 1; j < numTrees; j++) {
			likelihoods[j][n] -= meanLik[j];
			if (likelihoods[j][n] > maxLik)
				maxLik = likelihoods[j][n];
		}

	/* Calculate the distribution of deltas for this replicate */
		for (j = 0; j < numTrees; j++)
			likelihoods[j][n] = maxLik - likelihoods[j][n];
	}

	fprintf(stdout, "\tTree\tlnL\tdelta\tP(delta)\n");
	for (j = 0; j < numTrees; j++) {
		if (j == maxTree)
			fprintf(stdout, "ML");

		probTree = 0.0;
		for (n = 0; n < numReplicates; n++) {
			if (treeLikelihoods[j] <= likelihoods[j][n])
				probTree++;
		}
		probTree /= numReplicates;
		
		fprintf(stdout, "\t%d\t%lf\t%lf\t%lf\n", j+1, 
			bestLik - treeLikelihoods[j],
			treeLikelihoods[j], probTree);
	}
}

void PerformPARATest(TTree **trees, double *treeLikelihoods)
{
	double **likelihoods;
	double *meanLik;
	double maxLik, bestLik, probTree;
	long n, i, j, k, r, maxTree;
	char *P, *Q;
	
	meanLik = (double *)AllocMem(numTrees*sizeof(double), "meanLik", "PerformFULLTest", 1);

	likelihoods = (double **)AllocMem(numTrees*sizeof(double*), "likelihoods", "PerformFULLTest", 1);
	for (j = 0; j < numTrees; j++) {
		likelihoods[j] = (double *)AllocMem(numReplicates*sizeof(double), "likelihoods[]", "PerformFULLTest", 1);
	}
	
	/* Find the maximum likelihood tree */
	maxLik = treeLikelihoods[0];
	maxTree = 0;
	for (j = 1; j < numTrees; j++) {
		if (treeLikelihoods[j] > maxLik) {
			maxLik = treeLikelihoods[j];
			maxTree = j;
		}
	}

	bestLik = maxLik;

	/* Calculate the deltas for each tree */
	for (j = 0; j < numTrees; j++) {
		treeLikelihoods[j] = maxLik - treeLikelihoods[j];
		meanLik[j] = 0.0;
	}

	if (!quiet) {
		fprintf(stderr, "Generating bootstrap replicates...\n");
		InitProgressBar(numReplicates);
		DrawProgressBar();
	}
	
	/* Set up the simulation routines */
	SetupEvolve();
	
	/* For each bootstrap replicate... */
	for (n = 0; n < numReplicates; n++) {
		for (j = 0; j < numTrees; j++) {
		
			/* Evolve the sequences down the tree...*/
			EvolveSequences(trees[j]);
		
			/* Copy the sequences over to the main sequence array...*/
			for (i = 0; i < numTaxa; i++) {	
				P = sequences[i];
				Q = trees[j]->nodes[i].sequence;
				for (k = 0; k < numBases; k++) {
					*P = nucleotideTable[*Q];
					P++; Q++;
				}
			}
				
			CreatePatterns();
		
			SetTipLikelihoods(trees[j]);
			quiet = 1;
			SetupLikelihood(trees[j]);
			quiet = 0;

			TreeLik(trees[j]);
			likelihoods[j][n] = trees[j]->lnL;
			meanLik[j] += trees[j]->lnL;
		}
			
		if (!quiet)
			ProgressBar();
	}
	
	/* Centre the resampled likelihoods and find the maximum */
	for (j = 0; j < numTrees; j++)
		meanLik[j] /= numReplicates;

	for (n = 0; n < numReplicates; n++) {
		likelihoods[0][n] -= meanLik[0];
		maxLik = likelihoods[0][n];
		for (j = 1; j < numTrees; j++) {
			likelihoods[j][n] -= meanLik[j];
			if (likelihoods[j][n] > maxLik)
				maxLik = likelihoods[j][n];
		}

	/* Calculate the distribution of deltas for this replicate */
		for (j = 0; j < numTrees; j++)
			likelihoods[j][n] = maxLik - likelihoods[j][n];
	}

	fprintf(stdout, "\tTree\tlnL\tdelta\tP(delta)\n");
	for (j = 0; j < numTrees; j++) {
		if (j == maxTree)
			fprintf(stdout, "ML");

		probTree = 0.0;
		for (n = 0; n < numReplicates; n++) {
			if (treeLikelihoods[j] <= likelihoods[j][n])
				probTree++;
		}
		probTree /= numReplicates;
		
		fprintf(stdout, "\t%d\t%lf\t%lf\t%lf\n", j+1, 
			bestLik - treeLikelihoods[j],
			treeLikelihoods[j], probTree);
	}
}

int main(int argc, char **argv)
{
	int i, n, oldTaxa, oldBases;
	double lik;
	TTree *tree;
	TTree **trees = NULL;
	clock_t coreStart, totalStart;
	double coreSecs, totalSecs;
	double **siteLikelihoods = NULL;
	double *treeLikelihoods;
		
#ifdef __MWERKS__
	argc=ccommand(&argv);
#endif	

	setvbuf( stderr, NULL, _IOLBF, 512 );
	setvbuf( stdout, NULL, _IOLBF, 512 );

	totalMem=0;
 	
	totalStart = clock();
	
	SetSeed(time(NULL));
	
	ReadParams(argc, argv);

	ReadFileParams();
	
	AllocateMemory();
		
	ReadFile();
	
	if (gammaRates)
		DiscreteGamma(freqRate, siteRate, alpha, alpha, numCategories, 0);

	CreatePatterns();

	tree=CreateTree(0);
	AllocateTreeLikelihoods(tree);
		
	/*if (userTrees)*/
	numTrees=OpenTreeFile();
	numSets = numTrees;
	if (numTrees <= 1) {
		fprintf(stderr, "Need more than one tree to perform the SH test.\n\n");
		PrintUsage();
		exit(0);
	}
	
	ReadNextTree(tree);
	
	if (testKind != RELL) {
		trees = (TTree **)AllocMem(numTrees*sizeof(TTree*), "trees", "main", 1);
	}
	
	if (!quiet && verbose)
 		PrintVerbose(tree);

	SetTipLikelihoods(tree);
	SetupLikelihood(tree);
	 
	coreStart = clock();

	if (!quiet && !showProgress) {
		fprintf(stderr, "no.\tlnL");
		LabelFunc();
		if (estimateRates) {
			if (numCategories==2)
				fprintf(stderr, "\trate1\trate2");
			else
				fprintf(stderr, "\trate1\trate2\trate3");
		} else if (estimateAlpha) {
			fprintf(stderr, "\talpha");
		}
		fprintf(stderr, "\n");
	}
	
	if (testKind == RELL) 
		siteLikelihoods = (double **)AllocMem(numTrees*sizeof(double*), "siteLikelihoods", "main", 1);
	else if (testKind == PARA)
		AllocateTreeSequences(tree);
		
	treeLikelihoods = (double *)AllocMem(numTrees*sizeof(double), "treeLikelihoods", "main", 1);

	if (!quiet && showProgress) {
		InitProgressBar(numTrees);
		DrawProgressBar();
	}
		
	for (n=0; n<numTrees; n++) {
		TreeLik(tree);
		treeLikelihoods[n]=tree->lnL;
		
		if (!quiet && !showProgress) {
			fprintf(stderr, "%d\t%lf", n+1, treeLikelihoods[n]);
			
			FinishLikelihood(tree);
			
			if (estimateRates) {
				if (numCategories==2)
					fprintf(stderr, "\t%lf\t%lf", siteRate[0], siteRate[1]);
				else
					fprintf(stderr, "\t%lf\t%lf\t%lf", siteRate[0], siteRate[1], siteRate[2]);
			} else if (estimateAlpha) {
				fprintf(stderr, "\t%lf", alpha);
			}
		
			fprintf(stderr, "\n");
		}
		
		if (writeTrees) {
			WriteTree(stdout, tree, estimateClocks);
			fprintf(stdout, "\n");
		}

		if (testKind == RELL) {
			siteLikelihoods[n] = (double *)AllocMem(numBases*sizeof(double), "siteLikelihoods[n]", "main", 1);
			StoreSiteLik(siteLikelihoods[n]);
		} else if (testKind == PARA) {
			for (i=0; i<numCategories; i++)
				memcpy(tree->transMatrix[i], transMatrix[i], sizeof(double)*16*numBranches);
		}
				
		if (verboseSiteLik)
			PrintSiteLik(stdout, siteLikelihoods[n]);

		if (testKind != RELL) {
			trees[n] = tree;
			tree = NULL;
		}
	
		if (n<numTrees-1) {
			if (testKind != RELL) {
				tree=CreateTree(0);
				AllocateTreeLikelihoods(tree);
				if (testKind == PARA) {
					AllocateTreeSequences(tree);
				}
			}
					
			ReadNextTree(tree);
			SetTipLikelihoods(tree);
			quiet = 1;
			SetupLikelihood(tree);
			quiet = 0;
		}
		
		if (!quiet && showProgress)
			ProgressBar();
	}
	
	switch (testKind) {
		case RELL: PerformRELLTest(treeLikelihoods, siteLikelihoods); break;
		case FULL: PerformFULLTest(trees, treeLikelihoods); break;
		case PARA: PerformPARATest(trees, treeLikelihoods); break;
	}
	
	coreSecs = (double)(clock() - coreStart) / CLOCKS_PER_SEC;		
	
	CloseTreeFile();

	totalSecs = (double)(clock() - totalStart) / CLOCKS_PER_SEC;
	
	if (!quiet) {
		fprintf(stderr, "\nTime taken - core: %G, total: %G seconds\n", coreSecs, totalSecs);
	
		if (verboseMemory)
			fprintf(stderr, "Total memory used: %d\n", totalMem);
	}
	
	/* Could despose of all the memory here, but then the program is about to end... */
}



